go/gccgoexportdata: correctly handle archive files containing string tables

If the name of an archive member is longer than 16 bytes, an ELF archive
file contains a string table of file names under the name "//", and this
member has no mode.  Skip such members.

+ Test.

Change-Id: Ib10db1cc42816c9002433be6176240e490678560
Reviewed-on: https://go-review.googlesource.com/32973
Reviewed-by: Ian Lance Taylor <iant@golang.org>
diff --git a/go/gccgoexportdata/gccgoexportdata_test.go b/go/gccgoexportdata/gccgoexportdata_test.go
index 131f413..1aa114c 100644
--- a/go/gccgoexportdata/gccgoexportdata_test.go
+++ b/go/gccgoexportdata/gccgoexportdata_test.go
@@ -9,28 +9,59 @@
 )
 
 // Test ensures this package can read gccgo export data from the
-// .go_export section of an ELF file.
+// .go_export from a standalone ELF file or such a file in an archive
+// library.
+//
+// The testdata/{short,long}.a ELF archive files were produced by:
+//
+//   $ echo 'package foo; func F()' > foo.go
+//   $ gccgo -c -fgo-pkgpath blah foo.go
+//   $ objcopy -j .go_export foo.o foo.gox
+//   $ ar q short.a foo.gox
+//   $ objcopy -j .go_export foo.o name-longer-than-16-bytes.gox
+//   $ ar q long.a name-longer-than-16-bytes.gox
+//
+// The file long.a contains an archive string table.
+//
+// The errors.gox file (an ELF object file) comes from the toolchain's
+// standard library.
 func Test(t *testing.T) {
-	f, err := os.Open("testdata/errors.gox")
-	if err != nil {
-		t.Fatal(err)
-	}
-	defer f.Close()
-	r, err := gccgoexportdata.NewReader(f)
-	if err != nil {
-		t.Fatal(err)
-	}
-	imports := make(map[string]*types.Package)
-	pkg, err := gccgoexportdata.Read(r, nil, imports, "errors")
-	if err != nil {
+	for _, test := range []struct {
+		filename, path, member, wantType string
+	}{
+		{"testdata/errors.gox", "errors", "New", "func(text string) error"},
+		{"testdata/short.a", "short", "F", "func()"},
+		{"testdata/long.a", "long", "F", "func()"},
+	} {
+		t.Logf("filename = %s", test.filename)
+		f, err := os.Open(test.filename)
+		if err != nil {
+			t.Error(err)
+			continue
+		}
+		defer f.Close()
+		r, err := gccgoexportdata.NewReader(f)
+		if err != nil {
+			t.Error(err)
+			continue
+		}
 
-		t.Fatal(err)
-	}
+		imports := make(map[string]*types.Package)
+		pkg, err := gccgoexportdata.Read(r, nil, imports, test.path)
+		if err != nil {
+			t.Error(err)
+			continue
+		}
 
-	// Check type of errors.New.
-	got := pkg.Scope().Lookup("New").Type().String()
-	want := "func(text string) error"
-	if got != want {
-		t.Errorf("New.Type = %s, want %s", got, want)
+		// Check type of designated package member.
+		obj := pkg.Scope().Lookup(test.member)
+		if obj == nil {
+			t.Errorf("%s.%s not found", test.path, test.member)
+			continue
+		}
+		if obj.Type().String() != test.wantType {
+			t.Errorf("%s.%s.Type = %s, want %s",
+				test.path, test.member, obj.Type(), test.wantType)
+		}
 	}
 }
diff --git a/go/gccgoexportdata/testdata/long.a b/go/gccgoexportdata/testdata/long.a
new file mode 100644
index 0000000..c150447
--- /dev/null
+++ b/go/gccgoexportdata/testdata/long.a
Binary files differ
diff --git a/go/gccgoexportdata/testdata/short.a b/go/gccgoexportdata/testdata/short.a
new file mode 100644
index 0000000..3cd756e
--- /dev/null
+++ b/go/gccgoexportdata/testdata/short.a
Binary files differ
diff --git a/go/gccgoexportdata/gccgoexportdata.go b/go/gccgoexportdata/gccgoexportdata.go
index ca5f75f..30ed521 100644
--- a/go/gccgoexportdata/gccgoexportdata.go
+++ b/go/gccgoexportdata/gccgoexportdata.go
@@ -74,17 +74,14 @@
 	return sec.Open(), nil
 }
 
 // firstSection returns the contents of the first non-empty section of the archive file.
 func firstSection(a []byte) ([]byte, error) {
 	for len(a) >= 60 {
 		var hdr []byte
 		hdr, a = a[:60], a[60:]
 
-		modeStr := string(string(hdr[40:48]))
-		mode, err := strconv.Atoi(strings.TrimSpace(modeStr))
-		if err != nil {
-			return nil, fmt.Errorf("invalid mode: %q", modeStr)
-		}
+		name := strings.TrimSpace(string(hdr[:16]))
 
 		sizeStr := string(hdr[48:58])
 		size, err := strconv.Atoi(strings.TrimSpace(sizeStr))
@@ -92,16 +89,25 @@
 			return nil, fmt.Errorf("invalid size: %q", sizeStr)
 		}
 
-		var payload []byte
-		payload, a = a[:size], a[size:]
+		if len(a) < size {
+			return nil, fmt.Errorf("invalid section size: %d", size)
+		}
 
-		if mode == 0 {
-			continue // skip "/"
+		// The payload is padded to an even number of bytes.
+		var payload []byte
+		payload, a = a[:size], a[size+size&1:]
+
+		// Skip special files:
+		//   "/"           archive symbol table
+		//   "/SYM64/"     archive symbol table on e.g. s390x
+		//   "//"          archive string table (if any filename is >15 bytes)
+		if name == "/" || name == "/SYM64/" || name == "//" {
+			continue
 		}
 
 		return payload, nil
 	}
-	return nil, fmt.Errorf("archive has no non-empty sections")
+	return nil, fmt.Errorf("archive has no regular sections")
 }
 
 // Read reads export data from in, decodes it, and returns type
@@ -74,17 +74,14 @@
 }
 
-// firstSection returns the contents of the first non-empty section of the archive file.
+// firstSection returns the contents of the first regular file in an ELF
+// archive (http://www.sco.com/developers/devspecs/gabi41.pdf, §7.2).
 func firstSection(a []byte) ([]byte, error) {
 	for len(a) >= 60 {